"""Web links to aplustools"""
from aplustools.web.request import fetch, Result, BatchRequestHandler
from aplustools.web.utils import WebPage, url_validator

__all__ = ["fetch", "Result", "BatchRequestHandler", "WebPage", "url_validator"]
